import pytest
from pathlib import Path
from typing import Dict, Tuple
from lsprotocol.types import Position, Range
from pygls.workspace import Document

from vyper_lsp.ast import AST


@pytest.fixture
def ast():
    """Basic AST fixture without any source code loaded."""
    return AST()


@pytest.fixture
def vyper_document(tmp_path) -> Tuple[Document, Path]:
    """Create a temporary Vyper document for testing."""
    file_path = tmp_path / "test_contract.vy"
    file_path.touch()
    doc = Document(uri=str(file_path), source="")
    return doc, file_path


@pytest.fixture
def example_documents() -> Dict[str, str]:
    """Return the source code of example documents."""
    examples_dir = Path(__file__).parent.parent / "examples"
    return {
        "Foo.vy": (examples_dir / "Foo.vy").read_text(),
        "LiquidityGauge.vy": (examples_dir / "LiquidityGauge.vy").read_text(),
    }


@pytest.fixture
def struct_code() -> str:
    """Return a simple Vyper contract with a struct definition."""
    return """
struct Point:
    x: uint256
    y: uint256

my_point: Point

@external
def set_point(x: uint256, y: uint256):
    self.my_point = Point(x=x, y=y)

@external
@view
def get_point() -> Point:
    return self.my_point
"""


@pytest.fixture
def enum_code() -> str:
    """Return a simple Vyper contract with an enum definition."""
    return """
flag Color:
    RED
    GREEN
    BLUE

my_color: Color

@external
def set_color(color: Color):
    self.my_color = color

@external
@view
def get_color() -> Color:
    return self.my_color
"""


@pytest.fixture
def function_code() -> str:
    """Return a simple Vyper contract with internal and external functions."""
    return """
x: uint256
y: uint256

@internal
def _calculate(a: uint256, b: uint256) -> uint256:
    return a + b

@external
def calculate(a: uint256, b: uint256) -> uint256:
    return self._calculate(a, b)

@external
def set_values(a: uint256, b: uint256):
    self.x = a
    self.y = b
"""


@pytest.fixture
def position(line: int = 0, character: int = 0) -> Position:
    """Create a Position instance for testing."""
    return Position(line=line, character=character)


@pytest.fixture
def range_fixture(
    start_line: int = 0, start_char: int = 0, end_line: int = 0, end_char: int = 0
) -> Range:
    """Create a Range instance for testing."""
    return Range(
        start=Position(line=start_line, character=start_char),
        end=Position(line=end_line, character=end_char),
    )


@pytest.fixture
def ast_with_code(request, ast):
    """
    Parameterized fixture that loads code into an AST instance.

    Usage:
    @pytest.mark.parametrize('ast_with_code', ['struct_code', 'enum_code', 'function_code'], indirect=True)
    def test_something(ast_with_code):
        # ast_with_code has the specified code loaded
        ...
    """
    if hasattr(request, "param"):
        code_fixture = request.getfixturevalue(request.param)
        ast.build_ast(code_fixture)
    return ast



from typing import List
from vyper.ast import nodes
from lsprotocol.types import (
    CompletionItem,
    CompletionItemLabelDetails,
    CompletionList,
    CompletionParams,
)
from pygls.server import LanguageServer
from pygls.workspace import Document
from vyper_lsp.utils import (
    format_fn,
)
from vyper_lsp.ast import AST

# Available base types
UNSIGNED_INTEGER_TYPES = {f"uint{8*(i)}" for i in range(32, 0, -1)}
SIGNED_INTEGER_TYPES = {f"int{8*(i)}" for i in range(32, 0, -1)}
INTEGER_TYPES = UNSIGNED_INTEGER_TYPES | SIGNED_INTEGER_TYPES

BYTES_M_TYPES = {f"bytes{i}" for i in range(32, 0, -1)}
DECIMAL_TYPES = {"decimal"}

BASE_TYPES = list({"bool", "address"} | INTEGER_TYPES | BYTES_M_TYPES | DECIMAL_TYPES)

DECORATORS = ["payable", "nonpayable", "view", "pure", "external", "internal", "deploy"]


class CompletionHandler:
    def __init__(self, ast: AST):
        self.ast = ast

    def get_completions(
        self, ls: LanguageServer, params: CompletionParams
    ) -> CompletionList:
        document = ls.workspace.get_text_document(params.text_document.uri)
        return self._get_completions_in_doc(document, params)

    def _dot_completions_for_module(
        self, element: str, top_level_node=None, line: str = ""
    ) -> List[CompletionItem]:
        completions = []
        for name, fn in self.ast.imports[element].functions.items():
            doc_string = ""
            if getattr(fn.ast_def, "doc_string", False):
                doc_string = fn.ast_def.doc_string.value

            out = format_fn(fn)

            # NOTE: this just gets ignored by most editors
            # so we put the signature in the documentation string also
            completion_item_label_details = CompletionItemLabelDetails(detail=out)

            doc_string = f"{out}\n{doc_string}"

            show_external: bool = isinstance(
                top_level_node, nodes.ExportsDecl
            ) or line.startswith("exports:")
            show_internal_and_deploy: bool = isinstance(
                top_level_node, nodes.FunctionDef
            )

            if show_internal_and_deploy and (fn.is_internal or fn.is_deploy):
                completions.append(
                    CompletionItem(
                        label=name,
                        documentation=doc_string,
                        label_details=completion_item_label_details,
                    )
                )
            elif show_external and fn.is_external:
                completions.append(
                    CompletionItem(
                        label=name,
                        documentation=doc_string,
                        label_details=completion_item_label_details,
                    )
                )

        return completions

    def _dot_completions_for_element(
        self, element: str, top_level_node=None, line: str = ""
    ) -> List[CompletionItem]:
        completions = []
        if element == "self":
            for fn in self.ast.get_internal_functions():
                completions.append(CompletionItem(label=fn))
            # TODO: This should exclude constants and immutables
            for var in self.ast.get_state_variables():
                completions.append(CompletionItem(label=var))
        elif self.ast.imports and element in self.ast.imports.keys():
            completions = self._dot_completions_for_module(
                element, top_level_node=top_level_node, line=line
            )
        elif element in self.ast.flags:
            members = self.ast.flags[element]._flag_members
            for member in members.keys():
                completions.append(CompletionItem(label=member))

        if isinstance(top_level_node, nodes.FunctionDef):
            var_declarations = top_level_node.get_descendants(
                nodes.AnnAssign, filters={"target.id": element}
            )
            assert len(var_declarations) <= 1
            for vardecl in var_declarations:
                type_name = vardecl.annotation.id
                structt = self.ast.structs.get(type_name, None)
                if structt:
                    for member in structt.members:
                        completions.append(CompletionItem(label=member))

        return completions

    def _get_completions_in_doc(
        self, document: Document, params: CompletionParams
    ) -> CompletionList:
        items = []
        current_line = document.lines[params.position.line].strip()
        custom_types = self.ast.get_user_defined_types()

        no_completions = CompletionList(is_incomplete=False, items=[])

        if not params.context:
            return no_completions

        if params.context.trigger_character == ".":
            # get element before the dot
            # TODO: this could lead to bugs if we're not at EOL
            element = current_line.split(" ")[-1].split(".")[0]

            pos = params.position
            surrounding_node = self.ast.find_top_level_node_at_pos(pos)

            # internal + imported fns, state vars, and flags
            dot_completions = self._dot_completions_for_element(
                element, top_level_node=surrounding_node, line=current_line
            )
            if len(dot_completions) > 0:
                return CompletionList(is_incomplete=False, items=dot_completions)
            else:
                for attr in self.ast.get_attributes_for_symbol(element):
                    items.append(CompletionItem(label=attr))
            completions = CompletionList(is_incomplete=False, items=items)
            return completions

        if params.context.trigger_character == "@":
            for dec in DECORATORS:
                items.append(CompletionItem(label=dec))
            completions = CompletionList(is_incomplete=False, items=items)
            return completions

        if params.context.trigger_character == ":":
            # return empty_completions if colon isn't for a type annotation
            object_declaration_keywords = [
                "flag",
                "struct",
                "event",
                "enum",
                "interface",
                "def",
            ]
            if any(
                current_line.startswith(keyword)
                for keyword in object_declaration_keywords
            ):
                return no_completions

            for typ in custom_types + BASE_TYPES:
                items.append(CompletionItem(label=typ, insert_text=f" {typ}"))

            completions = CompletionList(is_incomplete=False, items=items)
            return completions

        if params.context.trigger_character == " ":
            if current_line[-1] == ":":
                for typ in custom_types + BASE_TYPES:
                    items.append(CompletionItem(label=typ))

                completions = CompletionList(is_incomplete=False, items=items)
                return completions

        return CompletionList(is_incomplete=False, items=[])



import pytest


@pytest.mark.parametrize("ast_with_code", ["struct_code"], indirect=True)
def test_struct_attributes(ast_with_code):
    """Test that struct attributes are correctly identified."""
    # The ast_with_code fixture has loaded struct_code
    ast = ast_with_code

    # Test struct detection
    assert ast.get_structs() == ["Point"]

    # Test struct fields
    assert ast.get_struct_fields("Point") == ["x", "y"]

    # Test references to the struct
    references = ast.find_nodes_referencing_struct("Point")
    assert (
        len(references) >= 3
    )  # Declaration, usage in set_point, return type in get_point


@pytest.mark.parametrize("ast_with_code", ["enum_code"], indirect=True)
def test_enum_attributes(ast_with_code):
    """Test that enum attributes are correctly identified."""
    # The ast_with_code fixture has loaded enum_code
    ast = ast_with_code

    # Test enum detection
    assert ast.get_enums() == ["Color"]

    # Test enum variants
    assert ast.get_enum_variants("Color") == ["RED", "GREEN", "BLUE"]

    # Test references to the enum
    references = ast.find_nodes_referencing_enum("Color")
    assert len(references) >= 2  # Variable declaration and return type in get_color


@pytest.mark.parametrize("ast_with_code", ["function_code"], indirect=True)
def test_function_attributes(ast_with_code):
    """Test that function attributes are correctly identified."""
    # The ast_with_code fixture has loaded function_code
    ast = ast_with_code

    # Test internal function detection
    assert ast.get_internal_functions() == ["_calculate"]

    # Test function references
    references = ast.find_nodes_referencing_internal_function("_calculate")
    assert len(references) == 1  # Referenced in calculate function

    # Test state variable references
    x_refs = ast.find_nodes_referencing_state_variable("x")
    assert len(x_refs) == 1  # Referenced in set_values

    y_refs = ast.find_nodes_referencing_state_variable("y")
    assert len(y_refs) == 1  # Referenced in set_values


@pytest.mark.parametrize(
    "code_fixture_name, expected_types",
    [("struct_code", ["Point"]), ("enum_code", ["Color"]), ("function_code", [])],
)
def test_user_defined_types(request, ast, code_fixture_name, expected_types):
    """Test user-defined types across different code samples."""
    # Load the code from the specified fixture
    code = request.getfixturevalue(code_fixture_name)
    ast.build_ast(code)

    # Verify the user-defined types match expectations
    assert sorted(ast.get_user_defined_types()) == sorted(expected_types)


def test_real_example_file(example_documents, ast):
    """Test using a real example file from the examples directory."""
    # Load the Foo.vy example
    foo_code = example_documents["Foo.vy"]
    ast.build_ast(foo_code)

    # Verify structs
    assert "Bar" in ast.get_structs()
    assert ast.get_struct_fields("Bar") == ["x"]

    # Verify flags
    assert "Roles" in ast.get_enums()
    assert sorted(ast.get_enum_variants("Roles")) == ["ADMIN", "USER"]

    # Verify constants
    assert "FEE" in ast.get_constants()

    # Verify functions
    assert ast.get_internal_functions() == ["bar"]



from vyper_lsp import utils


def test_get_word_at_cursor():
    text = "self.foo = 123"
    assert utils.get_word_at_cursor(text, 0) == "self"
    assert utils.get_word_at_cursor(text, 1) == "self"
    assert utils.get_word_at_cursor(text, 5) == "foo"
    assert utils.get_word_at_cursor(text, 12) == "123"

    text = "foo_bar = 123"
    assert utils.get_word_at_cursor(text, 0) == "foo_bar"
    assert utils.get_word_at_cursor(text, 4) == "foo_bar"


def test_get_expression_at_cursor():
    text = "self.foo = 123"
    assert utils.get_expression_at_cursor(text, 0) == "self.foo"
    assert utils.get_expression_at_cursor(text, 1) == "self.foo"
    assert utils.get_expression_at_cursor(text, 5) == "self.foo"
    assert utils.get_expression_at_cursor(text, 12) == "123"

    text = "foo_bar = self.baz (1,2,3)"
    assert utils.get_expression_at_cursor(text, 0) == "foo_bar"
    assert utils.get_expression_at_cursor(text, 4) == "foo_bar"
    assert utils.get_expression_at_cursor(text, 21) == "self.baz (1,2,3)"


def test_parse_fncall_expression():
    text = "self.foo()"
    assert utils.parse_fncall_expression(text) == ("self", "foo")

    text = "self.foo(self.bar())"
    assert utils.parse_fncall_expression(text) == ("self", "bar")

    text = "foobar"
    assert utils.parse_fncall_expression(text) is None



from lsprotocol.types import Position
from pygls.workspace import Document

from vyper_lsp.ast import AST
from vyper_lsp.navigation import ASTNavigator

import pytest


@pytest.fixture
def ast():
    ast = AST()
    return ast


@pytest.fixture
def doc(ast):
    doc = Document(uri="examples/Foo.vy")
    ast.build_ast(doc.source)
    return doc


@pytest.fixture
def navigator(ast):
    return ASTNavigator(ast)


def test_find_references_event_name(doc, navigator):
    pos = Position(line=2, character=7)
    references = navigator.find_references(doc, pos)
    assert len(references) == 3


def test_find_references_struct_name(doc, navigator):
    pos = Position(line=6, character=7)
    references = navigator.find_references(doc, pos)
    assert len(references) == 4


def test_find_references_enum_name(doc, navigator):
    pos = Position(line=9, character=7)
    references = navigator.find_references(doc, pos)
    assert len(references) == 5


def test_find_references_enum_variants(doc, navigator):
    pos = Position(line=10, character=7)
    references = navigator.find_references(doc, pos)
    assert len(references) == 1

    pos = Position(line=11, character=7)
    references = navigator.find_references(doc, pos)
    assert len(references) == 1


def test_find_references_function_name(doc, navigator):
    pos = Position(line=40, character=5)
    references = navigator.find_references(doc, pos)
    assert len(references) == 1


def test_find_references_storage_var(doc, navigator):
    pos = Position(line=13, character=0)
    references = navigator.find_references(doc, pos)
    assert len(references) == 3


def test_find_references_constant(doc, navigator):
    pos = Position(line=16, character=0)
    references = navigator.find_references(doc, pos)
    assert len(references) == 1

    pos = Position(line=60, character=0)
    references = navigator.find_references(doc, pos)
    assert len(references) == 2


def test_find_references_function_local_var(doc, navigator):
    pos = Position(line=20, character=5)
    references = navigator.find_references(doc, pos)
    assert len(references) == 1


def test_find_internal_fn_implementation(doc, navigator: ASTNavigator):
    pos = Position(line=35, character=17)
    implementation = navigator.find_implementation(doc, pos)
    assert implementation and implementation.start.line == 40


def test_find_interface_fn_implementation(doc, navigator: ASTNavigator):
    pos = Position(line=51, character=10)
    implementation = navigator.find_implementation(doc, pos)
    assert implementation and implementation.start.line == 57


def test_find_declaration_constant(doc, navigator: ASTNavigator):
    pos = Position(line=21, character=19)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 16


def test_find_declaration_struct(doc, navigator: ASTNavigator):
    pos = Position(line=22, character=7)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 6

    pos = Position(line=22, character=14)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 6


def test_find_declaration_enum(doc, navigator: ASTNavigator):
    pos = Position(line=25, character=7)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 9


def test_find_declaration_enum_variant(doc, navigator: ASTNavigator):
    # TODO: this currently jumps to the enum declaration, not the variant
    pos = Position(line=25, character=25)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 9


def test_find_declaration_event(doc, navigator: ASTNavigator):
    pos = Position(line=24, character=10)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 2


def test_find_declaration_function(doc, navigator: ASTNavigator):
    pos = Position(line=35, character=17)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 40


def test_find_declaration_storage_var(doc, navigator: ASTNavigator):
    pos = Position(line=26, character=9)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 13


def test_find_declaration_function_local_var(doc, navigator: ASTNavigator):
    pos = Position(line=26, character=13)
    declaration = navigator.find_declaration(doc, pos)
    assert declaration and declaration.start.line == 22


def test_find_implementation_variable_returns_none(doc, navigator: ASTNavigator):
    pos = Position(line=26, character=13)
    implementation = navigator.find_implementation(doc, pos)
    assert implementation is None



import logging
import string
import re
from pathlib import Path
from importlib.metadata import version
from typing import Optional, Tuple
from lsprotocol.types import Diagnostic, DiagnosticSeverity, Position, Range
from packaging.version import Version
from pygls.workspace import Document
from vyper.ast import VyperNode
from vyper.exceptions import VyperException
from vyper.compiler import FileInput

logger = logging.getLogger("vyper-lsp")


def get_installed_vyper_version():
    return Version(version("vyper"))


def get_source(filepath):
    base_path = Path(__file__).parent.parent
    filepath = base_path / filepath
    return filepath.read_text()


# detect if current line is a variable declaration
def is_var_declaration(line):
    # regex for variable declaration
    # should match lines starting with any identifier followed by a colon
    # like "foo: "
    reg = r"^\s*[a-zA-Z_][a-zA-Z0-9_]*\s*:"
    return bool(re.match(reg, line.strip()))


def is_attribute_access(line):
    # regex for attribute access
    # should match lines ending with a dot
    # like "foo."
    reg = r"\s*\.\s*$"
    return bool(re.match(reg, line.strip()))


_WORD_CHARS = string.ascii_letters + string.digits + "_"


# REVIEW: these get_.*_at_cursor helpers would benefit from having
# access to as much cursor information as possible (ex. line number),
# it could open up some possibilies when refactoring for performance


def get_word_at_cursor(sentence: str, cursor_index: int) -> str:
    start = cursor_index
    end = cursor_index

    # TODO: this could be a perf hotspot
    # Find the start of the word
    while start > 0 and sentence[start - 1] in _WORD_CHARS:
        start -= 1

    # Find the end of the word
    while end < len(sentence) and sentence[end] in _WORD_CHARS:
        end += 1

    # Extract the word
    word = sentence[start:end]

    return word


def _check_if_cursor_is_within_parenthesis(sentence: str, cursor_index: int) -> bool:
    # Find the nearest '(' before the cursor
    start = sentence[:cursor_index][::-1].find("(")
    if start != -1:
        start = cursor_index - start - 1

    # Find the nearest ')' after the cursor
    end = sentence[cursor_index:].find(")")
    if end != -1:
        end += cursor_index

    # Check if cursor is within a valid pair of parentheses
    if start != -1 and end != -1 and start < cursor_index < end:
        return True

    return False


def _get_entire_function_call(sentence: str, cursor_index: int) -> str:
    # Regex pattern to match function calls
    # This pattern looks for a word (function name), followed by optional spaces,
    # and then parentheses with anything inside.
    pattern = r"\b(?:\w+\.)*\w+\s*\([^)]*\)"

    # Find all matches in the sentence
    matches = [match for match in re.finditer(pattern, sentence)]

    # Find the match that contains the cursor
    for match in matches:
        if match.start() <= cursor_index <= match.end():
            return match.group()

    return ""  # Return an empty string if no match is found


def get_expression_at_cursor(sentence: str, cursor_index: int) -> str:
    if _check_if_cursor_is_within_parenthesis(sentence, cursor_index):
        return _get_entire_function_call(sentence, cursor_index)

    # does the same thing as get_word_at_cursor but includes . and [ and ] in the expression
    start = cursor_index
    end = cursor_index

    # Find the start of the word
    while start > 0 and sentence[start - 1] in _WORD_CHARS + ".[]":
        start -= 1

    # Find the end of the word
    while end < len(sentence) and sentence[end] in _WORD_CHARS + ".[]":
        end += 1

    # Extract the word
    word = sentence[start:end]

    return word


def extract_enum_name(line: str):
    m = re.match(r"enum\s+([a-zA-Z_][a-zA-Z0-9_]*)\s*:", line)
    if m:
        return m.group(1)
    return None


def range_from_node(node: VyperNode) -> Range:
    return Range(
        start=Position(line=node.lineno - 1, character=node.col_offset),
        end=Position(line=node.end_lineno - 1, character=node.end_col_offset),
    )


def range_from_exception(node: VyperException) -> Range:
    if getattr(node, "end_lineno", None) is None:
        return Range(
            start=Position(line=node.lineno - 1, character=node.col_offset),
            end=Position(line=node.lineno - 1, character=node.col_offset),
        )
    return Range(
        start=Position(line=node.lineno - 1, character=node.col_offset),
        end=Position(line=node.end_lineno - 1, character=node.end_col_offset),
    )


def create_diagnostic_warning(
    line_num: int, character_start: int, character_end: int, message: str
) -> Diagnostic:
    """
    Helper function to create a diagnostic object.

    :param line_num: The line number of the diagnostic.
    :param character_start: The starting character position of the diagnostic.
    :param character_end: The ending character position of the diagnostic.
    :param message: The diagnostic message.
    :return: A Diagnostic object.
    """
    return Diagnostic(
        range=Range(
            start=Position(line=line_num, character=character_start),
            end=Position(line=line_num, character=character_end),
        ),
        message=message,
        severity=DiagnosticSeverity.Warning,
    )


def diagnostic_from_exception(node: VyperException, message=None) -> Diagnostic:
    return Diagnostic(
        range=range_from_exception(node),
        message=message or str(node),
        severity=DiagnosticSeverity.Error,
    )


def document_to_fileinput(doc: Document) -> FileInput:
    path = Path(doc.uri.replace("file://", ""))
    return FileInput(0, path, path, doc.source)


def working_directory_for_document(doc: Document) -> Path:
    return Path(doc.uri.replace("file://", "")).parent


def escape_underscores(expression: str) -> str:
    return expression.replace("_", "\\_")


def format_fn(func) -> str:
    args = ", ".join([f"{arg.name}: _{arg.typ}_" for arg in func.arguments])
    return_value = f" -> _{func.return_type}_" if func.return_type is not None else ""
    mutability = func.mutability.value
    out = (
        f"def __{escape_underscores(func.name)}__({args}){return_value}: _{mutability}_"
    )
    return out


def parse_fncall_expression(expression: str) -> Optional[Tuple[str, str]]:
    # regex for matching 'module.function' or 'module.function(args)', not capturing args
    fncall_pattern = "(.*)\\.([^\\(]+)(?:\\(.*\\))?"

    if matches := re.match(fncall_pattern, expression):
        groups = matches.groups()
        module, fn = groups
        if "(" in module:
            module = module.split("(")[-1]
        return module, fn



import logging

from pygls.workspace import Document
from vyper.ast import FunctionDef
from typing import Optional

from lsprotocol.types import (
    ParameterInformation,
    SignatureHelp,
    SignatureHelpParams,
    SignatureInformation,
)
from vyper_lsp import utils
from vyper_lsp.ast import AST
from vyper_lsp.utils import get_expression_at_cursor

logger = logging.getLogger("vyper-lsp")


class SignatureHandler:
    def __init__(self, ast: AST):
        self.ast = ast

    def _handle_internal_fn_signature(
        self, current_line: str, fn_name: str
    ) -> Optional[SignatureHelp]:
        node = self.ast.find_function_declaration_node_for_name(fn_name)
        if not node:
            return None

        fn_name = node.name
        parameters = []

        fn_label = node.node_source_code.split(":\n")[0].removeprefix("def ")

        for arg in node.args.args:
            start_index = fn_label.find(arg.arg)
            end_index = start_index + len(arg.arg)
            parameters.append(
                ParameterInformation(label=(start_index, end_index), documentation=None)
            )
        active_parameter = current_line.split("(")[-1].count(",")
        return SignatureHelp(
            signatures=[
                SignatureInformation(
                    label=fn_label,
                    parameters=parameters,
                    documentation=None,
                    active_parameter=active_parameter or 0,
                )
            ],
            active_signature=0,
        )

    def _handle_imported_fn_signature(
        self, current_line: str, module: str, fn_name: str
    ) -> Optional[SignatureHelp]:
        if module in self.ast.imports:
            if fn := self.ast.imports[module].functions[fn_name]:
                logger.info(f"getting signature for {fn_name}")
                logger.info(fn.decl_node)
                node: FunctionDef = fn.decl_node
                label = node.node_source_code.split("def ")[1].split(":\n")[0]
                parameters = []
                for arg in node.args.args:
                    parameters.append(
                        ParameterInformation(label=arg.arg, documentation=None)
                    )
                active_parameter = current_line.split("(")[-1].count(",")
                return SignatureHelp(
                    signatures=[
                        SignatureInformation(
                            label=label,
                            parameters=parameters,
                            documentation=None,
                            active_parameter=active_parameter or 0,
                        )
                    ],
                    active_signature=0,
                )

    def signature_help(
        self, doc: Document, params: SignatureHelpParams
    ) -> Optional[SignatureHelp]:
        # TODO: Implement checking external functions, module functions, and interfaces
        current_line = doc.lines[params.position.line]
        expression = get_expression_at_cursor(
            current_line, params.position.character - 1
        )
        parsed = utils.parse_fncall_expression(expression)
        if parsed is None:
            return None
        module, fn_name = parsed

        # this returns for all external functions
        if module == "self":
            return self._handle_internal_fn_signature(current_line, fn_name)
        else:
            return self._handle_imported_fn_signature(current_line, module, fn_name)



from lsprotocol.types import (
    CompletionContext,
    CompletionParams,
    CompletionTriggerKind,
    Position,
    TextDocumentIdentifier,
)
from pygls.workspace import Document

from vyper_lsp.handlers.completion import CompletionHandler


def test_completion_internal_fn(ast):
    src = """
@internal
def foo():
    return

@external
def bar():
    self.foo()
"""
    ast.build_ast(src)

    src += """
@external
def baz():
    self.
"""

    doc = Document(uri="<inline source code>", source=src)
    pos = Position(line=11, character=7)
    context = CompletionContext(trigger_character=".", trigger_kind=2)
    params = CompletionParams(
        text_document=TextDocumentIdentifier(doc.uri), position=pos, context=context
    )

    analyzer = CompletionHandler(ast)
    completions = analyzer._get_completions_in_doc(doc, params)
    assert len(completions.items) == 1
    assert "foo" in [c.label for c in completions.items]


def test_completions_struct_members(ast):
    src = """
struct Foo:
    bar: uint256
    baz: uint256

@internal
def foo():
    return

@external
def bar():
    self.foo()

@external
def baz():
    g: Foo = Foo(bar=12, baz=13)
"""
    ast.build_ast(src)

    src += """
    g.
"""
    doc = Document(uri="<inline source code>", source=src)
    pos = Position(line=17, character=6)
    context = CompletionContext(trigger_character=".", trigger_kind=2)
    params = CompletionParams(
        text_document=TextDocumentIdentifier(uri=doc.uri), position=pos, context=context
    )

    analyzer = CompletionHandler(ast)
    completions = analyzer._get_completions_in_doc(doc, params)
    assert len(completions.items) == 2


def test_completions_enum_variant(ast):
    src = """
flag Foo:
    BAR
    BAZ

@internal
def foo():
    return

@external
def bar():
    self.foo()
"""
    ast.build_ast(src)

    src += """
@external
def baz():
    x: Foo = Foo.
"""

    doc = Document(uri="<inline source code>", source=src)
    pos = Position(line=15, character=18)
    context = CompletionContext(trigger_character=".", trigger_kind=2)
    params = CompletionParams(
        text_document=TextDocumentIdentifier(uri=doc.uri), position=pos, context=context
    )

    analyzer = CompletionHandler(ast)
    completions = analyzer._get_completions_in_doc(doc, params)
    assert len(completions.items) == 2
    assert "BAR" in [c.label for c in completions.items]
    assert "BAZ" in [c.label for c in completions.items]


def test_completion_fn_decorator(ast):
    src = """
@internal
def foo():
    return

@external
def bar():
    self.foo()
"""
    ast.build_ast(src)

    src += """
@
"""

    doc = Document(uri="<inline source code>", source=src)
    pos = Position(line=8, character=1)
    context = CompletionContext(
        trigger_character="@", trigger_kind=CompletionTriggerKind.TriggerCharacter
    )
    params = CompletionParams(
        text_document=TextDocumentIdentifier(uri=doc.uri), position=pos, context=context
    )

    analyzer = CompletionHandler(ast)
    completions = analyzer._get_completions_in_doc(doc, params)
    assert len(completions.items) == 7
    labels = [c.label for c in completions.items]
    assert "internal" in labels
    assert "external" in labels
    assert "payable" in labels
    assert "nonpayable" in labels
    assert "view" in labels
    assert "pure" in labels
    assert "deploy" in labels



import copy
import logging
from typing import Optional, List
from lsprotocol.types import Diagnostic, Position
from pygls.workspace import Document
from vyper.ast import VyperNode, nodes
from vyper.compiler import CompilerData
from vyper.compiler.input_bundle import FilesystemInputBundle
from vyper.compiler.phases import DEFAULT_CONTRACT_PATH, ModuleT
from vyper.semantics.types import StructT
from vyper.semantics.types.user import FlagT
from vyper.exceptions import VyperException
from vyper.cli.vyper_compile import get_search_paths
import warnings
import re

from vyper_lsp.utils import (
    create_diagnostic_warning,
    diagnostic_from_exception,
    working_directory_for_document,
    document_to_fileinput,
)

logger = logging.getLogger("vyper-lsp")


pattern_text = r"(.+) will be deprecated in a future release, use (.+) instead\."
deprecation_pattern = re.compile(pattern_text)


class AST:
    ast_data = None
    ast_data_annotated = None

    custom_type_node_types = (nodes.StructDef, nodes.FlagDef)

    # Module Data
    functions = {}
    variables = {}
    flags = {}
    structs = {}

    # Import Data
    imports = {}

    @classmethod
    def from_node(cls, node: VyperNode):
        ast = cls()
        ast.ast_data = node
        ast.ast_data_annotated = node
        return ast

    def _load_import_data(self):
        ast = self.ast_data_annotated
        if ast is None:
            return
        import_nodes = ast.get_descendants((nodes.ImportFrom, nodes.Import))
        node: nodes.ImportFrom | nodes.Import
        imports = {}
        for node in import_nodes:
            import_info = node._metadata["import_info"]
            module_t: ModuleT = import_info.typ.module_t
            alias = node._metadata["import_info"].alias
            imports[alias] = module_t

        self.imports = imports

    def _load_module_data(self):
        ast = self.ast_data_annotated
        if ast is None:
            return
        self.functions = ast._metadata["type"].functions
        self.variables = ast._metadata["type"].variables

        flagt_list = [
            FlagT.from_FlagDef(node) for node in ast._metadata["type"].flag_defs
        ]
        self.flags = {flagt.name: flagt for flagt in flagt_list}

        structt_list = [
            StructT.from_StructDef(node) for node in ast._metadata["type"].struct_defs
        ]
        self.structs = {structt.name: structt for structt in structt_list}

    def update_ast(self, doc: Document) -> List[Diagnostic]:
        diagnostics = self.build_ast(doc)
        return diagnostics

    def build_ast(self, doc: Document | str) -> List[Diagnostic]:
        if isinstance(doc, str):
            doc = Document(uri=str(DEFAULT_CONTRACT_PATH), source=doc)
        uri_parent_path = working_directory_for_document(doc)
        search_paths = get_search_paths([str(uri_parent_path)])
        fileinput = document_to_fileinput(doc)
        compiler_data = CompilerData(
            fileinput, input_bundle=FilesystemInputBundle(search_paths)
        )
        diagnostics = []
        replacements = {}
        warnings.simplefilter("always")
        with warnings.catch_warnings(record=True) as w:
            try:
                # unforunately we need this deep copy so the ast doesnt change
                # out from under us when folding stuff happens
                self.ast_data = copy.deepcopy(compiler_data.vyper_module)
                self.ast_data_annotated = compiler_data.annotated_vyper_module

                self._load_module_data()
                self._load_import_data()

            except VyperException as e:
                # make message string include class name
                message = f"{e.__class__.__name__}: {e}"
                if e.lineno is not None and e.col_offset is not None:
                    diagnostics.append(diagnostic_from_exception(e))
                if e.annotations:
                    for a in e.annotations:
                        diagnostics.append(
                            diagnostic_from_exception(a, message=message)
                        )

            for warning in w:
                m = deprecation_pattern.match(str(warning.message))
                if not m:
                    continue
                deprecated = m.group(1)
                replacement = m.group(2)
                replacements[deprecated] = replacement

        # Iterate over doc.lines and find all deprecated values
        for i, line in enumerate(doc.lines):
            for deprecated, replacement in replacements.items():
                for match in re.finditer(re.escape(deprecated), line):
                    character_start = match.start()
                    character_end = match.end()
                    diagnostic_message = (
                        f"{deprecated} is deprecated. Please use {replacement} instead."
                    )
                    diagnostics.append(
                        create_diagnostic_warning(
                            line_num=i,
                            character_start=character_start,
                            character_end=character_end,
                            message=diagnostic_message,
                        )
                    )

        return diagnostics

    @property
    def best_ast(self):
        if self.ast_data_annotated:
            return self.ast_data_annotated
        elif self.ast_data:
            return self.ast_data

        return None

    def get_descendants(self, *args, **kwargs):
        if self.best_ast is None:
            return []
        return self.best_ast.get_descendants(*args, **kwargs)

    def get_top_level_nodes(self, *args, **kwargs):
        if self.best_ast is None:
            return []
        return self.best_ast.get_children(*args, **kwargs)

    def get_enums(self) -> List[str]:
        # return [node.name for node in self.get_descendants(nodes.FlagDef)]
        return list(self.flags.keys())

    def get_structs(self) -> List[str]:
        # return [node.name for node in self.get_descendants(nodes.StructDef)]
        return list(self.structs.keys())

    def get_events(self) -> List[str]:
        return [node.name for node in self.get_descendants(nodes.EventDef)]

    def get_user_defined_types(self):
        return [node.name for node in self.get_descendants(self.custom_type_node_types)]

    def get_constants(self):
        # NOTE: Constants should be fetched from self.ast_data, they are missing
        # from self.ast_data_unfolded and self.ast_data_folded
        # NOTE: This may no longer be the case with the new AST format
        if self.ast_data is None:
            return []

        return [
            node.target.id
            for node in self.ast_data.get_children(
                nodes.VariableDecl, {"is_constant": True}
            )
        ]

    def get_enum_variants(self, enum: str):
        enum_node = self.find_type_declaration_node_for_name(enum)
        if enum_node is None:
            return []

        return [node.value.id for node in enum_node.get_children(nodes.Expr)]

    def get_struct_fields(self, struct: str):
        struct_node = self.find_type_declaration_node_for_name(struct)
        if struct_node is None:
            return []

        return [node.target.id for node in struct_node.get_children(nodes.AnnAssign)]

    def get_state_variables(self):
        # NOTE: The state variables should be fetched from self.ast_data, they are
        # missing from self.ast_data_unfolded and self.ast_data_folded when constants
        if self.ast_data is None:
            return []

        return [
            node.target.id for node in self.ast_data.get_descendants(nodes.VariableDecl)
        ]

    def get_internal_function_nodes(self):
        function_nodes = self.get_descendants(nodes.FunctionDef)
        internal_nodes = []

        for node in function_nodes:
            for decorator in node.decorator_list:
                if isinstance(decorator, nodes.Name) and decorator.id == "internal":
                    internal_nodes.append(node)

        return internal_nodes

    def get_internal_functions(self):
        internal_fn_names = [k for k, v in self.functions.items() if v.is_internal]
        return internal_fn_names

    def find_nodes_referencing_internal_function(self, function: str):
        return self.get_descendants(
            nodes.Call, {"func.attr": function, "func.value.id": "self"}
        )

    def find_nodes_referencing_state_variable(self, variable: str):
        return self.get_descendants(
            nodes.Attribute, {"value.id": "self", "attr": variable}
        )

    def find_nodes_referencing_constant(self, constant: str):
        name_nodes = self.get_descendants(nodes.Name, {"id": constant})
        return [
            node
            for node in name_nodes
            if not isinstance(node.get_ancestor(), nodes.VariableDecl)
        ]

    def get_attributes_for_symbol(self, symbol: str):
        node = self.find_type_declaration_node_for_name(symbol)
        if node is None:
            return []

        if isinstance(node, nodes.StructDef):
            return self.get_struct_fields(symbol)
        elif isinstance(node, nodes.FlagDef):
            return self.get_enum_variants(symbol)

        return []

    def find_function_declaration_node_for_name(
        self, function: str
    ) -> Optional[nodes.FunctionDef]:
        for node in self.get_descendants(nodes.FunctionDef):
            name_match = node.name == function
            not_interface_declaration = not isinstance(
                node.get_ancestor(), nodes.InterfaceDef
            )
            if name_match and not_interface_declaration:
                return node

        return None

    def find_state_variable_declaration_node_for_name(self, variable: str):
        # NOTE: The state variables should be fetched from self.ast_data, they are
        # missing from self.ast_data_unfolded and self.ast_data_folded when constants
        if self.ast_data is None:
            return None

        for node in self.ast_data.get_descendants(nodes.VariableDecl):
            if node.target.id == variable:
                return node

        return None

    def find_type_declaration_node_for_name(self, symbol: str):
        searchable_types = self.custom_type_node_types + (nodes.EventDef,)
        for node in self.get_descendants(searchable_types):
            if node.name == symbol:
                return node
            if isinstance(node, nodes.FlagDef):
                for variant in node.get_children(nodes.Expr):
                    if variant.value.id == symbol:
                        return variant

        return None

    def find_nodes_referencing_enum(self, enum: str):
        return_nodes = []

        for node in self.get_descendants(nodes.AnnAssign, {"annotation.id": enum}):
            return_nodes.append(node)
        for node in self.get_descendants(nodes.Attribute, {"value.id": enum}):
            return_nodes.append(node)
        for node in self.get_descendants(nodes.VariableDecl, {"annotation.id": enum}):
            return_nodes.append(node)
        for node in self.get_descendants(nodes.FunctionDef, {"returns.id": enum}):
            return_nodes.append(node)

        return return_nodes

    def find_nodes_referencing_enum_variant(self, enum: str, variant: str):
        return self.get_descendants(
            nodes.Attribute, {"attr": variant, "value.id": enum}
        )

    def find_nodes_referencing_struct(self, struct: str):
        return_nodes = []

        for node in self.get_descendants(nodes.AnnAssign, {"annotation.id": struct}):
            return_nodes.append(node)
        for node in self.get_descendants(nodes.Call, {"func.id": struct}):
            return_nodes.append(node)
        for node in self.get_descendants(nodes.VariableDecl, {"annotation.id": struct}):
            return_nodes.append(node)
        for node in self.get_descendants(nodes.FunctionDef, {"returns.id": struct}):
            return_nodes.append(node)

        return return_nodes

    def find_top_level_node_at_pos(self, pos: Position) -> Optional[VyperNode]:
        nodes = self.get_top_level_nodes()
        for node in nodes:
            if node.lineno <= pos.line and pos.line <= node.end_lineno:
                return node

        # return node with highest lineno if no node found
        if nodes:
            # sort
            nodes.sort(key=lambda x: x.lineno, reverse=True)
            return nodes[0]

        return None

    def find_nodes_referencing_symbol(self, symbol: str):
        # this only runs on subtrees
        return_nodes = []

        for node in self.get_descendants(nodes.Name, {"id": symbol}):
            parent = node.get_ancestor()
            if isinstance(parent, nodes.Dict):
                # skip struct key names
                if symbol not in [key.id for key in parent.keys]:
                    return_nodes.append(node)
            elif isinstance(parent, nodes.AnnAssign):
                if node.id == parent.target.id:
                    # lhs of variable declaration
                    continue
                else:
                    return_nodes.append(node)
            else:
                return_nodes.append(node)

        return return_nodes

    def find_node_declaring_symbol(self, symbol: str):
        for node in self.get_descendants((nodes.AnnAssign, nodes.VariableDecl)):
            if node.target.id == symbol:
                return node

        return None



from lsprotocol.types import Position, SignatureHelpParams, TextDocumentIdentifier
from pygls.workspace import Document
from vyper_lsp.ast import AST
from vyper_lsp.handlers.signatures import SignatureHandler
from vyper_lsp.handlers.hover import HoverHandler


def test_signature_help(ast: AST):
    src = """
@internal
def foo(x: int128, y: int128) -> int128:
    return x + y

@external
def bar():
    self.foo(1, 2)

@internal
def baz(x: int128) -> int128:
    return x

@external
def foobar():
    self.foo(self.baz(1), 2)
"""
    ast.build_ast(src)
    analyzer = SignatureHandler(ast)

    doc = Document(uri="<inline source code>", source=src)

    pos = Position(line=3, character=11)
    params = SignatureHelpParams(
        text_document=TextDocumentIdentifier(doc.uri), position=pos
    )
    sig_help = analyzer.signature_help(doc, params)
    assert sig_help is None

    pos = Position(line=7, character=13)
    params = SignatureHelpParams(
        text_document=TextDocumentIdentifier(doc.uri), position=pos
    )

    sig_help = analyzer.signature_help(doc, params)
    assert sig_help
    assert sig_help.active_signature == 0
    assert sig_help.signatures[0].active_parameter == 1
    assert sig_help.signatures[0].label == "foo(x: int128, y: int128) -> int128"

    pos = Position(line=15, character=22)
    params = SignatureHelpParams(
        text_document=TextDocumentIdentifier(doc.uri), position=pos
    )
    sig_help = analyzer.signature_help(doc, params)
    assert sig_help
    assert sig_help.active_signature == 0
    assert sig_help.signatures[0].active_parameter == 1
    assert sig_help.signatures[0].label == "baz(x: int128) -> int128"


def test_hover(ast: AST):
    src = """
@internal
def foo(
    x: int128,
    y: int128
) -> int128:
    return x + y

@external
def bar():
    self.foo(1, 2)

@internal
def noreturn(x: uint256):
    y: uint256 = x

@internal
def baz():
    self.noreturn(1)
"""
    ast.build_ast(src)

    doc = Document(uri="<inline source code>", source=src)

    pos = Position(line=10, character=11)

    handler = HoverHandler(ast)
    hover = handler.hover_info(doc, pos)
    assert hover
    assert (
        hover
        == """(Internal Function) def foo(
    x: int128,
    y: int128
) -> int128:"""
    )

    pos = Position(line=18, character=11)
    hover = handler.hover_info(doc, pos)
    assert hover
    assert hover == "(Internal Function) def noreturn(x: uint256):"



from lsprotocol.types import Position
from vyper_lsp.ast import AST


def test_get_constants(ast):
    src = """
x: constant(uint256) = 123
y: uint256
z: constant(bool) = True

@deploy
def __init__():
    self.y = x
        """
    ast.build_ast(src)
    assert ast.get_constants() == ["x", "z"]


def test_get_flags(ast):
    src = """
flag Foo:
    Bar
    Baz
    """
    ast.build_ast(src)
    assert ast.get_enums() == ["Foo"]


def test_get_flag_variants(ast):
    src = """
flag Foo:
    Bar
    Baz
    """
    ast.build_ast(src)
    assert ast.get_enum_variants("Foo") == ["Bar", "Baz"]
    assert ast.get_enum_variants("Bar") == []


def test_get_struct_fields(ast):
    src = """
struct Foo:
    bar: uint256
    baz: address
    """
    ast.build_ast(src)
    assert ast.get_struct_fields("Foo") == ["bar", "baz"]
    assert ast.get_struct_fields("Bar") == []


def test_get_events(ast):
    src = """
event Foo:
    arg1: indexed(uint256)
    arg2: indexed(address)
    arg3: uint256
    """
    ast.build_ast(src)
    assert ast.get_events() == ["Foo"]


def test_get_structs(ast):
    src = """
struct Foo:
    bar: uint256
    baz: address
    """
    ast.build_ast(src)
    assert ast.get_structs() == ["Foo"]


def test_get_user_defined_types(ast):
    src = """
struct Foo:
    bar: uint256
    baz: address

event FooEvent:
    arg1: indexed(uint256)
    arg2: indexed(address)

flag FooFlag:
    Bar
    Baz
    """
    ast.build_ast(src)
    assert ast.get_user_defined_types() == ["Foo", "FooFlag"]


def test_get_state_variables(ast):
    src = """
x: uint256
y: address
z: bool
        """
    ast.build_ast(src)
    assert ast.get_state_variables() == ["x", "y", "z"]


def test_get_internal_functions(ast):
    src = """
@internal
def foo():
    pass

@external
def bar():
    pass
        """
    ast.build_ast(src)
    assert ast.get_internal_functions() == ["foo"]


def test_find_nodes_referencing_internal_function(ast):
    src = """
@internal
def foo():
    pass

@external
def bar():
    self.foo()
        """
    ast.build_ast(src)
    references = ast.find_nodes_referencing_internal_function("foo")
    assert len(references) == 1
    assert (
        references[0].lineno == 8
    )  # line number of self.foo(), counting first newline


def test_find_nodes_referencing_state_variable(ast):
    src = """
x: uint256
y: address
z: bool

@external
def foo():
    self.x = 123
    self.y = msg.sender
    self.z = True
        """
    ast.build_ast(src)
    references = ast.find_nodes_referencing_state_variable("x")
    assert len(references) == 1
    assert (
        references[0].lineno == 8
    )  # line number of self.x = 123, counting first newline

    references = ast.find_nodes_referencing_state_variable("y")
    assert len(references) == 1
    assert (
        references[0].lineno == 9
    )  # line number of self.y = msg.sender, counting first newline

    references = ast.find_nodes_referencing_state_variable("z")
    assert len(references) == 1
    assert (
        references[0].lineno == 10
    )  # line number of self.z = True, counting first newline


def test_find_nodes_referencing_constant(ast):
    src = """
x: constant(uint256) = 123
y: uint256
z: constant(bool) = True

@external
def foo():
    self.y = x
        """
    ast.build_ast(src)
    references = ast.find_nodes_referencing_constant("x")
    assert len(references) == 1
    assert (
        references[0].lineno == 8
    )  # line number of self.y = x, counting first newline

    references = ast.find_nodes_referencing_constant("z")
    assert len(references) == 0


def test_find_nodes_referencing_constant_folded(ast):
    src = """
x: constant(uint256) = 10 ** 2
y: uint256

@external
def foo():
    self.y = x
        """
    ast.build_ast(src)
    references = ast.find_nodes_referencing_constant("x")
    assert len(references) == 1
    assert (
        references[0].lineno == 7
    )  # line number of self.y = x, counting first newline


def test_find_nodes_referencing_flag(ast):
    src = """
flag Foo:
    Bar
    Baz

@external
def foo():
    x: Foo = Foo.Bar
    y: Foo = Foo.Baz

@external
def bar() -> Foo:
    return Foo.Bar
        """
    ast.build_ast(src)
    references = ast.find_nodes_referencing_enum("Foo")
    assert len(references) == 6
    assert references[0].lineno == 8  # line number of self.Bar, counting first newline
    assert references[3].lineno == 9  # line number of self.Baz, counting first newline


def test_find_nodes_referencing_flag_variant(ast):
    src = """
flag Foo:
    Bar
    Baz

@external
def foo():
    x: Foo = Foo.Bar
    y: Foo = Foo.Baz
        """
    ast.build_ast(src)
    references = ast.find_nodes_referencing_enum_variant("Foo", "Bar")
    assert len(references) == 1
    assert references[0].lineno == 8  # line number of self.Bar, counting first newline

    references = ast.find_nodes_referencing_enum_variant("Foo", "Baz")
    assert len(references) == 1
    assert references[0].lineno == 9  # line number of self.Baz, counting first newline


def test_find_nodes_referencing_struct(ast):
    src = """
struct Foo:
    bar: uint256
    baz: address

@external
def foo():
    x: Foo = Foo(bar=123, baz=msg.sender)

@external
def bar() -> Foo:
    return Foo(bar=123, baz=msg.sender)
        """
    ast.build_ast(src)
    references = ast.find_nodes_referencing_struct("Foo")
    assert len(references) == 4
    assert (
        references[0].lineno == 8
    )  # line number of x: Foo = Foo, counting first newline


def test_find_nodes_referencing_symbol(ast: AST):
    src = """
@internal
def foo() -> uint256:
    x: uint256 = 123
    y: uint256 = x
    return y

@external
def bar():
    self.foo()
        """
    ast.build_ast(src)
    functiondef_node = ast.get_internal_function_nodes()[0]
    fn_ast = AST.from_node(functiondef_node)
    references = fn_ast.find_nodes_referencing_symbol("x")
    assert len(references) == 1
    assert (
        references[0].lineno == 5
    )  # line number of y: uint256 = x, counting first newline


def test_get_attributes_for_symbol(ast):
    src = """
struct Foo:
    bar: uint256
    baz: address

flag Bar:
    Baz

@external
def foo():
    x: Foo = Foo(bar=123, baz=msg.sender)
        """
    ast.build_ast(src)
    assert ast.get_attributes_for_symbol("Foo") == ["bar", "baz"]
    assert ast.get_attributes_for_symbol("Bar") == ["Baz"]
    assert ast.get_attributes_for_symbol("Baz") == []


def test_find_function_declaration_node_for_name(ast):
    src = """
@internal
def foo() -> uint256:
    x: uint256 = 123
    y: uint256 = x
    return y

@external
def bar():
    self.foo()
        """
    ast.build_ast(src)
    assert (
        ast.find_function_declaration_node_for_name("foo").lineno == 3
    )  # line number of def foo(), counting first newline
    assert ast.find_function_declaration_node_for_name("baz") is None


def test_find_state_variable_declaration_node_for_name(ast):
    src = """
x: uint256
y: address
z: bool

@external
def foo():
    self.x = 123
    self.y = msg.sender
    self.z = True
        """
    ast.build_ast(src)
    assert (
        ast.find_state_variable_declaration_node_for_name("x").lineno == 2
    )  # line number of x: uint256, counting first newline
    assert (
        ast.find_state_variable_declaration_node_for_name("y").lineno == 3
    )  # line number of y: address, counting first newline
    assert (
        ast.find_state_variable_declaration_node_for_name("z").lineno == 4
    )  # line number of z: bool, counting first newline
    assert ast.find_state_variable_declaration_node_for_name("baz") is None


def test_find_type_declaration_node_for_name(ast):
    src = """
struct Foo:
    bar: uint256
    baz: address

flag Bar:
    Baz

@external
def foo():
    x: Foo = Foo(bar=123, baz=msg.sender)
        """
    ast.build_ast(src)
    assert (
        ast.find_type_declaration_node_for_name("Foo").lineno == 2
    )  # line number of struct Foo, counting first newline
    assert (
        ast.find_type_declaration_node_for_name("Bar").lineno == 6
    )  # line number of flag Bar, counting first newline
    assert (
        ast.find_type_declaration_node_for_name("Baz").lineno == 7
    )  # line number of Baz, counting first newline
    assert ast.find_type_declaration_node_for_name("baz") is None


def test_find_top_level_node_at_position(ast):
    src = """
x: uint256
y: address
z: bool

flag Foo:
    Bar
    Baz

@external
def foo():
    self.x = 123
    self.y = msg.sender
    self.z = True
        """
    ast.build_ast(src)
    pos: Position = Position(line=13, character=0)
    assert (
        ast.find_top_level_node_at_pos(pos).lineno == 11
    )  # line number of self.x = 123, counting first newline

    pos: Position = Position(line=8, character=0)
    assert (
        ast.find_top_level_node_at_pos(pos).lineno == 6
    )  # line number of flag Foo, counting first newline


def test_find_node_declaring_symbol(ast):
    src = """
x: uint256
y: address

@external
def foo():
    self.x = 123
    self.y = msg.sender
        """
    ast.build_ast(src)
    assert (
        ast.find_node_declaring_symbol("x").lineno == 2
    )  # line number of x: uint256, counting first newline
    assert (
        ast.find_node_declaring_symbol("y").lineno == 3
    )  # line number of y: address, counting first newline


def test_ast_no_data_returns_empty_and_none(ast: AST):
    ast.ast_data = None
    ast.ast_data_annotated = None

    assert ast.get_constants() == []
    assert ast.get_enums() == []
    assert ast.get_enum_variants("Foo") == []
    assert ast.get_events() == []
    assert ast.get_structs() == []
    assert ast.get_user_defined_types() == []
    assert ast.get_state_variables() == []
    assert ast.get_internal_functions() == []
    assert ast.get_struct_fields("Foo") == []
    assert ast.get_internal_function_nodes() == []
    assert ast.find_nodes_referencing_internal_function("foo") == []
    assert ast.find_nodes_referencing_state_variable("x") == []
    assert ast.find_nodes_referencing_constant("x") == []
    assert ast.find_nodes_referencing_enum("Foo") == []
    assert ast.find_nodes_referencing_enum_variant("Foo", "Bar") == []
    assert ast.find_nodes_referencing_struct("Foo") == []
    assert ast.find_nodes_referencing_symbol("x") == []
    assert ast.get_attributes_for_symbol("Foo") == []
    assert ast.find_function_declaration_node_for_name("foo") is None
    assert ast.find_state_variable_declaration_node_for_name("x") is None
    assert ast.find_type_declaration_node_for_name("Foo") is None
    assert ast.find_top_level_node_at_pos(Position(line=0, character=0)) is None
    assert ast.find_node_declaring_symbol("x") is None



import time
from vyper_lsp.debounce import Debouncer  # Import Debouncer from your module


def test_debounce():
    result = []

    def test_function(arg):
        result.append(arg)

    debouncer = Debouncer(wait=0.5)
    debounced_func = debouncer.debounce(test_function)

    debounced_func("first call")
    time.sleep(0.2)  # Sleep for less than the debounce period
    debounced_func("second call")
    time.sleep(
        0.6
    )  # Sleep for more than the debounce period to allow the function to execute

    assert result == ["second call"]



from lsprotocol.types import Position
from pygls.workspace import Document

from vyper_lsp.ast import AST


def test_document_fixture(vyper_document, struct_code):
    """Test that the document fixture can be used to create a document with source code."""
    doc, file_path = vyper_document

    # Update the document with the struct code
    doc = Document(uri=str(file_path), source=struct_code)

    # Use the document to test AST functionality
    ast = AST()
    diagnostics = ast.build_ast(doc)

    # Verify no diagnostics were reported
    assert len(diagnostics) == 0

    # Verify the expected structs were found
    assert ast.get_structs() == ["Point"]


def test_document_diagnostics(vyper_document):
    """Test that malformed code produces diagnostics."""
    doc, file_path = vyper_document

    # Invalid Vyper code with a syntax error
    invalid_code = """
struct Point:
    x: uint256
    y uint256  # Missing colon

@external
def set_point(x: uint256, y: uint256):
    self.my_point = Point(x=x, y=y)
"""

    # Update the document with the invalid code
    doc = Document(uri=str(file_path), source=invalid_code)

    # Use the document to test AST functionality
    ast = AST()
    diagnostics = ast.build_ast(doc)

    # Verify diagnostics were reported
    assert len(diagnostics) > 0


def test_example_document_loading(example_documents, ast):
    """Test loading different example documents."""
    # Test with Foo.vy which we know is compatible
    ast.build_ast(example_documents["Foo.vy"])
    assert "Bar" in ast.get_structs()
    assert "Roles" in ast.get_enums()

    # Note: We're not testing LiquidityGauge.vy as it may have version
    # compatibility issues with the current Vyper version


def test_find_top_level_node_with_position(vyper_document, function_code):
    """Test finding a top-level node at a given position."""
    doc, file_path = vyper_document

    # Update the document with the function code
    doc = Document(uri=str(file_path), source=function_code)

    # Use the document to test AST functionality
    ast = AST()
    ast.build_ast(doc)

    # Position in the _calculate function (internal)
    internal_pos = Position(line=6, character=5)  # Line with "return a + b"
    internal_node = ast.find_top_level_node_at_pos(internal_pos)
    assert internal_node is not None
    assert hasattr(internal_node, "name")
    assert internal_node.name == "_calculate"

    # Position in the calculate function (external)
    external_pos = Position(
        line=10, character=5
    )  # Line with "return self._calculate(a, b)"
    external_node = ast.find_top_level_node_at_pos(external_pos)
    assert external_node is not None
    assert hasattr(external_node, "name")
    assert external_node.name == "calculate"



import logging
import re
from typing import Optional
from packaging.version import Version
from lsprotocol.types import (
    Position,
)
from pygls.workspace import Document
from vyper.ast import nodes
from vyper_lsp.ast import AST
from vyper_lsp.utils import (
    get_expression_at_cursor,
    get_word_at_cursor,
)

pattern_text = r"(.+) is deprecated\. Please use `(.+)` instead\."
deprecation_pattern = re.compile(pattern_text)

min_vyper_version = Version("0.4.0")

logger = logging.getLogger("vyper-lsp")


class HoverHandler:
    def __init__(self, ast: AST) -> None:
        self.ast = ast

    def _format_fn_signature(self, node: nodes.FunctionDef) -> str:
        pattern = r"def\s+(\w+)\((?:[^()]|\n)*\)(?:\s*->\s*[\w\[\], \n]+)?:"
        match = re.search(pattern, node.node_source_code, re.MULTILINE)
        if match:
            function_def = match.group()
            return f"(Internal Function) {function_def}"

    def _is_internal_fn(self, expression: str):
        if not expression.startswith("self."):
            return False
        fn_name = expression.split("self.")[-1]
        return fn_name in self.ast.functions and self.ast.functions[fn_name].is_internal

    def _is_state_var(self, expression: str):
        if not expression.startswith("self."):
            return False
        var_name = expression.split("self.")[-1]
        return var_name in self.ast.variables

    def hover_info(self, doc: Document, pos: Position) -> Optional[str]:
        if len(doc.lines) < pos.line:
            return None

        og_line = doc.lines[pos.line]
        word = get_word_at_cursor(og_line, pos.character)
        full_word = get_expression_at_cursor(og_line, pos.character)

        if self._is_internal_fn(full_word):
            node = self.ast.find_function_declaration_node_for_name(word)
            return node and self._format_fn_signature(node)

        if self._is_state_var(full_word):
            node = self.ast.find_state_variable_declaration_node_for_name(word)
            if not node:
                return None
            variable_type = node.annotation.id
            return f"(State Variable) **{word}** : **{variable_type}**"

        if word in self.ast.get_structs():
            node = self.ast.find_type_declaration_node_for_name(word)
            return node and f"(Struct) **{word}**"

        if word in self.ast.get_enums():
            node = self.ast.find_type_declaration_node_for_name(word)
            return node and f"(Enum) **{word}**"

        if word in self.ast.get_events():
            node = self.ast.find_type_declaration_node_for_name(word)
            return node and f"(Event) **{word}**"

        if word in self.ast.get_constants():
            node = self.ast.find_state_variable_declaration_node_for_name(word)
            if not node:
                return None

            variable_type = node.annotation.id
            return f"(Constant) **{word}** : **{variable_type}**"

        return None
